use std::{io::Read, path::PathBuf};

use based::{
    page::search::Search,
    request::{
        api::GeneratedPager, assets::DataResponse, respond_json, RequestContext, StringResponse,
    },
};
use maud::{html, PreEscaped};
use rocket::{get, State};

pub mod component;
use component::*;
use serde_json::json;

use webarc::{
    ai::{generate_embedding, remove_data_urls, EmbedStore, SearchResult},
    archive::{extract_domains, WebsiteArchive},
    conf::get_config,
    render_page,
};

const SEARCH_BAR_STYLE: &str = "w-full px-4 mb-4 py-2 text-white bg-black border-2 border-neon-blue placeholder-neon-blue focus:ring-2 focus:ring-neon-pink focus:outline-none font-mono text-lg";

/// Get the favicon of a domain
#[get("/favicon/<domain>")]
pub async fn favicon_route(domain: &str) -> Option<DataResponse> {
    let mut buf = Vec::new();
    std::fs::File::open(std::path::Path::new("./favicon").join(domain))
        .ok()?
        .read_to_end(&mut buf)
        .ok()?;

    // TODO : Default favicon

    Some(DataResponse::new(
        buf,
        "image/x-icon".to_string(),
        Some(60 * 60 * 24 * 5),
    ))
}

/// Websites Overview
#[get("/")]
pub async fn index(ctx: RequestContext, arc: &State<WebsiteArchive>) -> StringResponse {
    let websites = arc.domains();

    let content = html! {
        div class="container mx-auto p-4" {

            div class="mb-4" {
                input type="search" name="query" placeholder="Search..." class=(SEARCH_BAR_STYLE)
                hx-get=("/vector_search")
                hx-target="#website_grid" hx-push-url="true" hx-swap="outerHTML" {};
            };


        div id="website_grid" {
            h1 class="text-5xl font-bold text-center mb-10" { "Websites" };
            div class="grid grid-cols-2 sm:grid-cols-3 lg:grid-cols-5 xl:grid-cols-6 2xl:grid-cols-8 gap-6" {
                @for site in websites {
                    a href=(format!("/d/{site}")) class="bg-neutral-900 shadow-md rounded-lg hover:bg-neutral-800 bg-gray-1 hover:cursor-pointer transition-all duration-300 flex flex-col items-center justify-center aspect-square max-w-60" {
                        div class="bg-blue-500 text-white rounded-full p-4" {
                            (favicon(&site))
                        };
                        p class="mt-4 text-base font-medium" { (site) };
                    };
                };
            };
        };
        }
    };

    render_page(content, ctx).await
}

/// Overview on <domain> / <path>
#[get("/d/<domain>/<paths..>")]
pub async fn domain_info_route(
    ctx: RequestContext,
    domain: &str,
    paths: PathBuf,
    arc: &State<WebsiteArchive>,
) -> StringResponse {
    let domain = arc.get_domain(domain);
    let document = domain.path(paths.to_str().unwrap());
    let versions: Vec<String> = document.versions();
    let (path_entries, is_doc) = domain.paths(paths.to_str().unwrap());
    let path_seperations: Vec<&str> = paths.to_str().unwrap().split('/').collect();

    let domains = extract_domains(&document.render_local(None).await.unwrap_or_default());

    let content = html! {
        h2 class="text-xl font-bold mb-4 flex items-center" {
            img class="p-2" src=(format!("/favicon/{}", &domain.name)) {};
            a href=(format!("/d/{}", &domain.name)) { (domain.name) };
            (slash_seperator())
            (gen_path_header(path_seperations, &domain.name, true))
        };

        @if !versions.is_empty() {
            div class="max-w-md mx-auto p-4 bg-neutral-900 rounded-lg shadow-md mb-8" {
                h3 class="font-bold mb-2" { "Saved Versions:" };
                @for (index, ver) in versions.into_iter().enumerate() {
                    a href=(format!("/s/{}/{}?time={}", &domain.name, paths.to_str().unwrap(), ver)) class="text-xs text-blue-400" {

                        @if index == 0 {
                        (format!("Latest ({ver})"))
                        } @else {
                            (ver)
                        }
                    };
                }
            };
        }

        @if path_entries.is_empty() && !is_doc {
            div class="max-w-md mx-auto p-4 bg-neutral-900 rounded-lg shadow-md" {
                p class="p-4 text-xl" { "Nothing found" };
            }
        }

        @if !path_entries.is_empty() {
        div class="max-w-md mx-auto p-4 bg-neutral-900 rounded-lg shadow-md" {
            h3 class="font-bold mb-2" { "Paths:" };
            ul class="space-y-2" {
                @for path in path_entries {
                    a href=(path.url()) class="flex items-center gap-2 p-3 border bg-neutral-800 rounded hover:shadow-lg transition" {
                        (arrow_icon("blue"))

                        span class="font-medium" { (path.path()) };
                    };
                };
            };
        };
    };

    @if !domains.is_empty() {
        div class="max-w-md mx-auto p-4 bg-neutral-900 rounded-lg shadow-md" {
            h3 class="font-bold mb-2" { "Domains linked on this page:" };
            ul class="space-y-2 p-4" {
                @for domain in domains {
                    a href=(format!("/d/{domain}")) class="flex items-center gap-2 p-3 border bg-neutral-800 rounded hover:shadow-lg transition" {
                        (favicon(&domain));
                        span class="font-medium" { (domain) };
                    };
                };
            };
        };
    };
    };

    render_page(content, ctx).await
}

#[get("/txt/<domain>/<path..>?<time>&<no_data_urls>")]
pub async fn render_txt_website(
    domain: &str,
    path: PathBuf,
    time: Option<&str>,
    no_data_urls: Option<&str>,
    arc: &State<WebsiteArchive>,
) -> Option<String> {
    let document = arc.get_domain(domain).path(path.to_str().unwrap());

    let mut content = document
        .render_local(time.map(|time| time.to_string()))
        .await?;

    if no_data_urls.is_some() {
        content = remove_data_urls(&content);
    }

    Some(html2md::parse_html(&content))
}

/// Return archived version of `domain` / `path` at `time`
#[get("/s/<domain>/<path..>?<time>")]
pub async fn render_website(
    domain: &str,
    path: PathBuf,
    time: Option<&str>,
    arc: &State<WebsiteArchive>,
) -> Option<DataResponse> {
    let document = arc.get_domain(domain).path(path.to_str().unwrap());

    let content = document
        .render_local(time.map(|time| time.to_string()))
        .await;

    if let Some(content) = content {
        return Some(DataResponse::new(
            content.as_bytes().to_vec(),
            "text/html".to_string(),
            Some(60 * 60 * 24),
        ));
    } else if get_config().DOWNLOAD_ON_DEMAND && time.is_none() {
        arc.archive_url(&format!("https://{domain}/{}", path.to_str().unwrap()))
            .await;

        let content = document.render_local(None).await?;

        return Some(DataResponse::new(
            content.as_bytes().to_vec(),
            "text/html".to_string(),
            Some(60 * 60 * 24),
        ));
    }

    None
}

pub fn gen_search_element(x: &SearchResult) -> PreEscaped<String> {
    html! {
        div class="text-xl font-bold mt-4 p-4 flex items-center w-full max-w-4xl max-h-40 mx-auto bg-neutral-800 shadow-md rounded-lg overflow-hidden border border-neutral-900 hover:cursor-pointer"
        hx-get=(format!("/d/{}/{}", x.domain, x.path))
        hx-target="#main_content" hx-push-url="true" hx-swap="innerHTML"
        {
                (favicon(&x.domain))
                a { (x.domain) };
                (slash_seperator());
                (gen_path_header(x.path.split('/').collect(), &x.domain, false));
                p class="font-bold p-2 text-stone-400" { (format!("{:.2} % [{} matches]", x.similarity() * 100.0, x.chunks.len())) };
        };
    }
}

#[get("/vector_search?<query>&<page>")]
pub async fn vector_search(
    query: Option<&str>,
    page: Option<i64>,
    ctx: RequestContext,
) -> Option<StringResponse> {
    get_config().ai.as_ref()?;

    let page = page.unwrap_or(1);

    // Search
    let search =
        Search::new("/vector_search".to_string()).search_class(SEARCH_BAR_STYLE.to_string());

    if let Some(query) = query {
        // If we have query
        let real_query = query.trim_end_matches(".json");

        // Search Results
        let vector = pgvector::Vector::from(generate_embedding(real_query.to_string()).await?);

        let results = GeneratedPager::new(
            |input, offset, limit| {
                Box::pin(async move {
                    EmbedStore::search_vector(&input, limit as i64, offset as i64).await
                })
            },
            1500,
        )
        .pager(page as u64, vector)
        .await;

        // API Route
        if query.ends_with(".json") {
            // TODO : Better search API
            return Some(respond_json(&json!(&results.page(page as u64))));
        }

        let content = search.build_response(&ctx, results, page, real_query, gen_search_element);

        return Some(render_page(content, ctx).await);
    }

    // Return new search site
    let content = search.build("", html! {});
    Some(render_page(content, ctx).await)
}
